; ints module - "Nice" interrupt handlers
;
; 1. Exception handlers
; 2. Auto-generate default interrupt handlers. These tell you what
;    interrupt number it is, as a debugging aid.
;
; Notes:
;	Exception 8 (double fault) and 10 (invalid TSS) should have task
;	gates, so a known machine state is entered when these exceptions are
;	triggered.
;

	extern	patch_int
	extern	print, print_hex, putc
	extern	print_num

	%include "boot.mac"
	%include "odb.mac"
	%include "memory.inc"


	SECTION .text
init:	callproc print, msg_init
	call	generate
	callproc patch_int,	6, exc6_handler
	callproc patch_int,	12, exc12_handler
	callproc patch_int,	13, exc13_handler
	ret



generate:
	pusha
	mov	edi, IDT
	mov	ebx, [CODE_end]
	mov	ecx, 48		; Generate handlers for first 48 interrupts
	xor	edx, edx	; interrupt #
.loop
	;; Generate IDT entry
	mov	eax, ebx
	mov	[edi], ax
	shr	eax, 16
	mov	[edi+2], word 8	; flat_code selector
	mov	[edi+4], word 0x8E00; flags: INT and PRESENT
	mov	[edi+6], ax
	add	edi, 8

	;; Generate machine code for interrupt handler
	mov	[ebx], byte 0x68; push (interrupt #)
	mov	[ebx+1], edx
	inc	edx
	mov	[ebx+5], byte 0xE9; jmp default_int_handler
	mov	eax, default_int_handler; compute offset
	sub	eax, ebx
	sub	eax, 10
	mov	[ebx+6], eax
	add	ebx, 10		; 10 bytes of machine code
	loop	.loop
	mov	[CODE_end], ebx
	popa
	ret


default_int_handler:
	callproc print,		int_msg
	callproc print_num,	[esp]
	cr
	add	esp, 4
	iretd



exc6_handler:
	callproc print,		exc6_msg
	mov	eax, [esp]
	callproc print_hex,	eax
	cr
	sti
	jmp short $

exc12_handler:
	callproc print,		exc12_msg
	sti
	jmp short $

exc13_handler:
	callproc print,		exc13_msg
	callproc print_hex, dword [esp]
	callproc print_hex, dword [esp+4]
	cr
	sti
	jmp short $



	SECTION .data
msg_init	db	"++ Generic interrupt handlers",0
int_msg		db	10," * Interrupt ",0
exc6_msg	db	10," * EXC6: Invalid opcode",10,0
exc12_msg	db	10," * EXC12: Stack fault",10,0
exc13_msg	db	10," * EXC13: GPF",10,0
