; screen.asm - Direct screen output
;
; Copyright (C) 1999 Thomas Novelli
;	Retro is distributed under the terms of the GNU LGPL, version 2 or
;	greater. (For the text of the license, see the file COPYING.LIB)
;
; Progress:
;	1/24/99 Wrote putc, tputc, print, scroll_line, update_cursor, clear,
;		cr, space
;
;	3/19/99 Little improvements.. TAB, BS..
;
;	3/28/99 Obsoleted by new video routines in the kernel, which use a
;	more efficient cursor handling mechanism.
;
; To do:
;	More terminal handling stuff in tputc
;	Could use a "case statement equivalent".. tables?
;	Yeah, have an array of characters at es:edi. 'repne scasb'. If
;	found, use ECX as an index into an address table to call the proper
;	subroutine.
;

%include "kernel.mac"
%include "macros.mac"
%include "screen.mac"

	SECTION .text

;Segment descriptors (quick hack, this should be more automatic)
;flat_code	EQU	0x08
;flat_data	EQU	0x10

;WTF? EQU and ABSOLUTE are generating RDF IMPORT records. More NASM bugs.
;Fine, I'll use %define instead.
%define flat_code	0x08
%define flat_data	0x10
%define x		0x450
%define y		0x451


init:
	print	msg
	ret



	global ph	;shortcut :) gotta type this alot
ph:
	global _print_hex
_print_hex:
;
; Print a number in hexadecimal
;
; Usage:
;	push	<the number>
;	call	_print_hex
;	add	esp, 4
;
	push	ebp
	mov	ebp, esp
	push	eax
	push	ebx
	push	ecx
	push	edx

	mov	edx, [ebp+8]
        mov     ecx, 8
.print_digit:
        rol     edx,4
        mov     ax,0E0Fh
        and     al,dl
        add     al,90h
        daa
        adc     al,40h
        daa
	putc	eax
        loop    .print_digit

	pop	edx
	pop	ecx
	pop	ebx
	pop	eax
	leave
        ret


	global cr
cr:
	tputc 10
	ret

	global space
space:
	putc 32
	ret

	global p		;Another shortcut..
p:
	global _print
_print:
;
; Print a null-terminated string
;
; Usage:
;	push	<string's address>
;	call	_print
;	add	esp, 4
;
	push	ebp
	mov	ebp, esp
	push	eax
	push	ebx

	xor	eax, eax
	mov	ebx, [ebp+8]
.loop
	mov	al, [ebx]
	or	al, al
	jz	.done
	tputc	eax
	inc	ebx
	jmp short .loop
.done
	pop	ebx
	pop	eax
	leave
	ret


	global _tputc
_tputc:
;
; "TTY putc" - prints a character, interpreting special characters.
;
; Done: CR, LF, ^L, NULL, BS, TAB
; Todo: BELL, DEL
;
	push	ebp
	mov	ebp, esp
	push	eax
	push	ebx

	mov	al, [ebp+8]		;Fetch char & test it

	or	al, al			;NULL: Ignore
	jz	.end
	cmp	al, 13			;CR: Carriage return
	je	.cr

	cmp	al, 10			;LF: Carriage return
	jne	.not_lf
.cr
	mov	[x], byte 0		;X=0
	mov	ah, [y]
	inc	ah			;Y++
	mov	bh, [rows]		;Past bottom of screen?
	dec	bh
	cmp	ah, bh
	jbe	.noscroll
	call	scroll_line
	dec	ah
.noscroll
	mov	[y], ah
	call	update_cursor
	jmp short .end
.not_lf

	cmp	al, 12			;^L: Clear screen
	jne	.not_clear
	call	clear
	jmp short .end
.not_clear
	cmp	al, 8			;BS: Backspace
	jne	.not_bs
	dec	byte [x]
	call	update_cursor
	jmp short .end
.not_bs
	cmp	al, 9			;TAB
	jne	.not_tab
	mov	al, [x]
	shr	al, 3	;Round down to nearest 8
	shl	al, 3
	add	al, 8	;Tab over another 8
	mov	[x], al
	call update_cursor
	jmp short .end
.not_tab
	putc	[ebp+8]
.end
	pop	ebx
	pop	eax
	leave
	ret


	global _putc
_putc:
;
; Prints a character to the screen, adjusting the cursor position.
; Uses the BIOS's cursor position.
; Doesn't interpret any control characters (like CR/LF)
;
	push	ebp
	mov	ebp, esp
	mpush	eax, ebx, edx

	movzx	eax, byte [y]		;Coordinates...
	xor	edx, edx
	mul	dword [cols]
	shl	eax, 1
	movzx	ebx, byte [x]
	shl	ebx, 1
	add	eax, ebx
	add	eax, [buf]		;EAX = memory location to write

	mov	bh, [attr]
	mov	bl, [ebp+8]		;Print the character
	mov	[eax], bx

	mov	al, [x]			;Adjust cursor position
	inc	al			;AL = x
	mov	bl, [cols]		;Past right margin?
	dec	bl
	cmp	al, bl
	jbe	.nowrap

	xor	al, al			;Wrap to next line
	mov	ah, [y]			;AH = y
	inc	ah
	mov	bh, [rows]		;Past bottom of screen?
	dec	bh
	cmp	ah, bh
	jbe	.noscroll
	call	scroll_line
	dec	ah
.noscroll
	mov	[y], ah
.nowrap
	mov	[x], al
	call	update_cursor

	mpop	eax, ebx, edx
	leave
	ret


scroll_line:
;
; Scroll down 1 line
;
	push	eax
	push	ecx
	push	edx
	push	esi
	push	edi
	mov	ax, flat_data
	mov	es, ax

	mov	esi, [cols]		;Source = 2nd line
	shl	esi, 1
	add	esi, [buf]

	mov	edi, [buf]		;Destination = Top line

	mov	eax, [rows]
	dec	eax
	xor	edx, edx
	mul	dword [cols]
	shl	eax, 1
	mov	ecx, eax		;Count = (rows-1) * cols * 2

	cld				;Move the string
	rep movsb

	mov	ecx, [cols]		;Clear last line
	mov	ax, 0x0700		;attr/char to clear to
	rep	stosw

	pop	esi
	pop	edi
	pop	edx
	pop	ecx
	pop	eax
	ret


	global update_cursor
update_cursor:
;
; Physically move the cursor to [x,y], using port I/O
;
; Index port: 0x3d4
; Data port:  0x3d5
;
	push	eax
	push	edx

	mov	al, 0x0f		;Index: cursor pos (low byte)
	mov	dx, 0x3d4
	out	dx, al

	movzx	ax, [y]
	xor	dx, dx
	mul	word [cols]
	movzx	dx, byte [x]
	add	ax, dx			;AX = *linear* cursor position

	mov	dx, 0x3d5
	out	dx, al			;Write low byte

	mov	al, 0x0e		;Index: cursor pos (high byte)
	mov	dx, 0x3d4
	out	dx, al

	mov	al, ah
	mov	dx, 0x3d5
	out	dx, al			;Write high byte

	pop	edx
	pop	eax
	ret


	global clear
clear:
;
; Clear the screen
;
	push	eax
	push	ecx
	push	edx
	push	edi
	mov	ax, flat_data
	mov	es, ax

	mov	edi, [buf]		;Destination

	mov	eax, [rows]
	xor	edx, edx
	mul	dword [cols]
	mov	ecx, eax		;Count = rows * cols
	mov	ah, [attr]		;attr/char to clear to
	xor	al, al
	rep	stosw

	mov	[x], word 0		;Set X and Y to 0
	call	update_cursor

	pop	edi
	pop	edx
	pop	ecx
	pop	eax
	ret


	SECTION .data

	global attr
buf	dd	0xB8000			;Video buffer location
attr	db	0x07			;Current color attribute
rows	dd	25			;Screen size
cols	dd	80

msg	db	"Screen module loaded",10,0


;	ABSOLUTE 0x450			;BIOS data
;x:	resb	1
;y:	resb	1

